/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is bookmarktools.
 *
 * The Initial Developer of the Original Code is
 * Maik Gollnick <maik.gollnick@gmx.de>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
	Simple page skipper to display all search results as pages.
*/
var SkipPage = {
	lastPattern : null,
	chunk : 0,
	max : 0,
	chunkSize : 25,
	
	next : null,
	previous : null,
	pageLabel0 : null,
	pageLabel1 : null,
	page : null,
	to : null,
	container : null,
	
	/*
		Init all object entries.
	*/
	init : function(next, previous, pageLabel0, pageLabel1, page, to, container){
		this.next = next;
		this.previous = previous;
		this.pageLabel0 = pageLabel0;
		this.pageLabel1 = pageLabel1;
		this.page = page;
		this.to = to;
		this.container = container;
	},
	
	/*
		Set chunk size (default 25).
	*/
	setChunkSize : function(chunkSize){
		this.chunkSize = chunkSize;
	},
	
	/*
		Performs the search process.
	*/
	doSearch : function(pattern){
		this.lastPattern = pattern;
		this.chunk = 0;
		
		try{
			this.max = Index._search(this.lastPattern, this.container, this.chunk, this.chunkSize);
		}catch(e){
			alert(e);
		}
		
		if(this.max != 0){
			this.disableView(false);
			this.updatePage();
		}else {
			this.reset();
		}
	},
	
	/*
		Disables\ Enables page view.
	*/
	disableView : function(disable){
		this.previous.disabled = disable;
		this.page.disabled = disable;
		this.to.disabled = disable;
		this.pageLabel0.disabled = disable;
		this.pageLabel1.disabled = disable;
		this.next.disabled = disable;
	},
	
	/*
		Resets page view.
	*/
	reset : function(){
		this.pageLabel0.value = 0;
		this.pageLabel1.value = 0;
		this.chunk = 0;
		this.max = 0;
		this.lastPattern = null;
		this.disableView(true);
	},

	/*
		Skip to previous page.
	*/
	previousPage : function(){
		if(this.chunk - 1 >= 0){
			this.chunk--;
			this.max = Index._search(this.lastPattern, this.container, this.chunk, this.chunkSize);
			this.updatePage();
		}
	},

	/*
		Skip to next page.
	*/
	nextPage : function(){
		if((this.chunk + 1) * this.chunkSize < this.max){
			this.chunk++;
			this.max = Index._search(this.lastPattern, this.container, this.chunk, this.chunkSize);
			this.updatePage();
		}
	},

	/*
		Update page labels.
	*/
	updatePage : function(){
		var from = (this.chunk * this.chunkSize) + 1;
		var to = (((this.chunk + 1) * this.chunkSize) < this.max ? ((this.chunk + 1) * this.chunkSize) : this.max);
		this.pageLabel0.value = from;
		this.pageLabel1.value = to;
	}
};
