/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is bookmarktools.
 *
 * The Initial Developer of the Original Code is
 * Maik Gollnick <maik.gollnick@gmx.de>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function Logger(){

	var LEVEL = {INFO : 'INFO', WARN : 'WARN', FATAL : 'FATAL', DEBUG : 'DEBUG'};

	var file;

	this.init = function(filename){
		file = BookmarkTools.utils.fileio.getFileObject();
		file.append("cl_index");
		file.append("log");
		file.append(filename);

		if(!file.exists())
            file.createUnique(Ci.nsIFile.NORMAL_FILE_TYPE, 0666);
	};

	var writeToFile = function(value){
		BookmarkTools.utils.fileio.writeLineToFile(file, value);
	};
	
	var format = function(level, message, object, callee){
		var on = getObjectName(object);
		var fn = getFunctionName(object, callee);

		return level + ' - ' + on + '.' + fn + StringUtils.TAB + ':' + StringUtils.TAB + message + StringUtils.LF;
	};
	
	var getFunctionName = function(object, callee){
		for(var i in object){
			var member = object[i];
			if(member == callee) return i;
		}
		return 'unknown';
	};
	
	var getObjectName = function(object){
		var c = object.constructor.toString();
		return c.substring(c.indexOf(' '), c.indexOf('('));
	};
	
	this.info = function(message, object, callee){
		writeToFile(format(LEVEL.INFO, message, object, callee));
	};
	
	this.warn = function(message, object, callee){
		writeToFile(format(LEVEL.WARN, message, object, callee));
	};
	
	this.fatal = function(message, object, callee){
		writeToFile(format(LEVEL.FATAL, message, object, callee));
	};
	
	this.debug = function(message, object, callee){
		writeToFile(format(LEVEL.DEBUG, message, object, callee));
	};

}

function HTTPRequest(){
	this.method = "GET";
	this.uri = "";
	this.async = true;
	this.timeout = 10000;
	
	var httpRequest = null;
	var timeoutID = -1;
	var self = this;
	
	var getMimeType = function(str){
		if(!str || str == null) return null;
		
		var match = str.match(/(application|text)\/[a-zA-Z0-9-+.]+/gi);
		return match ? match[0] : null;
	};
	
	var getCharSet = function(str){
		if(!str || str == null) return null;
		
		var match = str.match(/charset=[a-zA-Z0-9-_]+/gi);
		return match ? match[0] : null;
	};
	
	var getMetaTags = function(str){
		if(!str || str == null) return null;
		
		var match = str.match(/<meta\/?[^>]+>/gi);
		return match ? match : null;
	};
	
	var getHost = function(str){
		var tmp = str.split('/');
		return tmp[0]+"//"+tmp[2];
	};
	
	var getMCList = function(str, uri){
		var mcList = [];

		try {
			var hrefs = str.match(/href[\s]*=[\s]*["'].+["']/gi);
			var host = getHost(uri);
		
			while(hrefs.length != 0)
				mcList.push(StringUtils.trim(host + hrefs.shift().replace(/href[\s]*=[\s]*["'][\s]*|["']/gi, "")));
		}catch(e){
			return [];
		}
		
		return mcList;
	};
	
	var abortReq = function(){
		if(httpRequest){
			httpRequest.onerror = function(){};
			httpRequest.onreadystatechange = function(){};
			httpRequest.abort();
			httpRequest = null;
			self.onError(-400, "Request timeout");
		}
	};
	
	var cancelTimeout = function(){
		if(timeoutID != -1){
			clearTimeout(timeoutID);
			timeoutID = -1;
		}
	};
	
	this.close = function(){
		if(httpRequest){
			httpRequest.onerror = function(){};
			httpRequest.onreadystatechange = function(){};
			httpRequest.abort();
			httpRequest = null;
		}
	};
	
	this.send = function(uri){

		if(!StringUtils.isURL(uri)){
			self.onError(-401, "Bad uri");
			return;
		}
		
		httpRequest = new XMLHttpRequest();
		httpRequest.mozBackgroundRequest = true;
		
		timeoutID = setTimeout(abortReq, this.timeout);
		
		httpRequest.onerror = function(){
			if(httpRequest.status && httpRequest.statusText)
				self.onError(httpRequest.status, httpRequest.statusText);
			else
				self.onError(-402, "Unknown error");
		};
		
		httpRequest.onreadystatechange = function(){

			if(httpRequest.readyState == 4){

				cancelTimeout();

				var mimeType = null, charSet = null, contentType = null;
				switch(httpRequest.status){
					case 0 :
						mimeType = "text/plain";
						charSet = "charset=us-ascii"; 
						self.onComplete(mimeType, charSet, httpRequest.responseText);
						return;
					case 200:
						try{
							contentType = httpRequest.getResponseHeader("Content-Type");
							
							if(contentType && contentType != null){
								charSet = getCharSet(contentType);
								mimeType = getMimeType(contentType);
							}
								
							if(charSet == null || mimeType == null){
								var metaTags = getMetaTags(httpRequest.responseText);
								if(metaTags){
									for(var i = 0, l = metaTags.length ;i < l; i++){
										var tag = metaTags[i];
										charSet = charSet == null ? getCharSet(tag) : charSet;
										mimeType = mimeType == null ? getMimeType(tag) : mimeType;
										if(charSet != null && mimeType != null) break;
									}
								}
							}
							
						}catch(e){
							alert(e);
						}
				
						self.onComplete(mimeType, charSet, httpRequest.responseText);
						return;
					case 300: 
						self.onRedirect(
							httpRequest.status, 
							getMCList(httpRequest.responseText, uri)
						); 
						return;
					case 301: 
						self.onRedirect(
							httpRequest.status, 
							[httpRequest.getResponseHeader("Location")]
						); 
						return;
					case 302: 
						self.onRedirect(
							httpRequest.status, 
							[httpRequest.getResponseHeader("Location")]
						); 
						return;
					case 304: 
						self.onRedirect(
							httpRequest.status
						); 
						return;
					default:
						if(httpRequest.status && httpRequest.status >= 400 && httpRequest.status < 600){
							self.onError(httpRequest.status, httpRequest.statusText);
							return;
						}
						self.onError(-404, "Not implemented: (status: " + httpRequest.status + ")");
				}
			}
			
		};
	
		httpRequest.open(this.method, uri, this.async);
		//httpRequest.overrideMimeType("text/xml; charset=UTF-8");
		httpRequest.send(null);
	};
	
	this.onComplete = function(mimeType, charSet, responseText){}; //only 200 and 0
	this.onRedirect = function(status, locations){}; //300/301/302/304
	//unknown/timeout/bad uri/not implemented/no content type/4xx/5xx
	//unknown mime type
	this.onError = function(status, statusText){}; 
	
}

/*
	Web crawler based on XMLHTTPRequest.
*/
var Crawler = {

	bookmarks : null,
	redirections : null,
	
	isAborted : false,
	timerID : -1,

	req: null,
	logger : null,
	
	init : function(){
		this.redirections = [];
		this.bookmarks = [];
		
		this.logger = new Logger(this, arguments);
		this.logger.init("crawler.log");
	},
	
	_init : function(){
		this.isAborted = false;
		this.redirections = null;
		this.bookmarks = null;
	},
	
	// overwrite this onXYZ functions for event handling
	onAbort : function(){},
	onDone : function(){},
	
	onError : function(status, statusText, bookmark){},
	onComplete : function(responseText, mimeType, charSet, bookmark){},
	
	setBookmarks : function(bookmarks){
		this.bookmarks = bookmarks;
	},
	
	_abort : function(){
		this.isAborted = true;
	},

	_build : function(){
		this.buildProcess();
	},
 
	buildProcess : function(){

		this.closeRequest();

		if(!this.isAborted){
			
			var bookmark = this.bookmarks.shift();
			
			if(bookmark){
				this.logger.info(Book.getURI(bookmark), this, arguments.callee);
				this.sendRequest(bookmark);
			}else {
				this.logger.info("DONE.", this, arguments.callee);
				this.onDone();
				this._init();
			}
			
		}else {
			this.logger.info("ABORTED.", this, arguments.callee);
			this.onAbort();
			this._init();
		}
	},

	/*
	 *	If no mimeType is available try to find a valid type dependent on the urls
	 *	ending. Otherwise return the given mimeType or in case of errors the default
	 *	type 'text/plain'.
	 *
	 *	@pre	mimeType is valid or not defined
	 *	@post	mimeType is valid && lowercase
	 */
	panicModeMimeType : function(mimeType, bookmark){
		//TODO: try to extract mimeType from DOCTYPE or namespace attribute
		try {
			if(!mimeType || mimeType == null){
				var uri = Book.getURI(bookmark);
				var match = uri.match(/\.[a-zA-Z]+$/gi);
				match = match ? match[0].replace('.', StringUtils.EMPTY) : null;
				if(match != null){
					match = match.toLowerCase();
					switch(match){
						case "xml": return "text/xml";
						case "txt": return "text/plain";
						case "htm": return "text/html";
						case "html": return "text/html";
						default: break;
					}
				}
				return "text/plain"; 
			}
			return mimeType.toLowerCase();
		} catch(e){
			return "text/plain"; 
		}
	},
	
	/*
	 *	If no charSet is available try to find the initial line <?xml ... ?> with encoding
	 *	attribute. Otherwise return the given charSet or in case of errors the default
	 *	type 'charset=us-ascii'.
	 *
	 *	@pre	mimeType != 'undefined' && mimeType != null && mimeType is lowercase
	 *	@post	charSet is set  && lowercase
	 */
	panicModeCharSet : function(charSet, mimeType, responseText){
		try {
			if(!charSet || charSet == null){
				if(mimeType == "text/xml" || mimeType == "application/xhtml+xml" || mimeType == "application/xml"){
					var match = responseText.match(/encoding[\s]*=[\s]*"[a-zA-Z0-9-]*"/gi);
					match = match ? match[0] : null;
					if(match != null){
						match = match.toLowerCase();
						return match.replace("encoding", "charset");
					}
				}
				return "charset=us-ascii";
			}
			return charSet.toLowerCase();
		} catch(e){
			return "charset=us-ascii"; 
		}
	},
	
	closeRequest : function(){
		if(this.req != null){
			this.req.close();
			this.req = null;
		}
	},
	
	sendRequest : function(bookmark){

		this.closeRequest();
		this.req = new HTTPRequest();

		this.req.onComplete = function(mimeType, charSet, responseText){
			try{
				mimeType = Crawler.panicModeMimeType(mimeType, bookmark);
				charSet = Crawler.panicModeCharSet(charSet, mimeType, responseText);

				if(Encoding.supportedMimeType(mimeType) && Encoding.supportedCharSet(charSet)){
					Crawler.finalize(responseText, mimeType, charSet, bookmark);
				}else {
					Crawler.onError(-403, "Unsupported content type(" + mimeType + ", " + charSet + ")", bookmark);
					Crawler.buildProcess();
				}
			}catch(e){alert(e);}
		};

		this.req.onError = function(status, statusText){
			Crawler.onError(status, statusText, bookmark);
			Crawler.buildProcess();
		};
		
		this.req.onRedirect = function(status, locations){
			Crawler.handleRedirect(status, locations, bookmark);
		};

		this.req.send(Book.getURI(bookmark));
	},

	redirectProcess : function(bookmark){
		
		this.closeRequest();
		if(!this.isAborted){
			
			var uri = this.redirections.shift();
			var size = this.redirections.length;
			
			if(uri){
				this.sendRedirect(uri, size, bookmark);
			}else {
				this.onError(-406, 'Unable to redirect: ' + Book.getURI(bookmark), bookmark);
				this.buildProcess();
			}
			
		}else {
			this.onAbort();
			this._init();
		}
	},
	
	sendRedirect : function(uri, size, bookmark){
	
		this.closeRequest();
		this.req = new HTTPRequest();
		
		this.req.onComplete = function(mimeType, charSet, responseText){
			try{
				mimeType = Crawler.panicModeMimeType(mimeType, bookmark);
				charSet = Crawler.panicModeCharSet(charSet, mimeType, responseText);

				if(Encoding.supportedMimeType(mimeType) && Encoding.supportedCharSet(charSet)){
					Crawler.finalize(responseText, mimeType, charSet, bookmark);
				}else {
					Crawler.onError(-403, "Unsupported content type(" + mimeType + ", " + charSet + ")", bookmark);
					Crawler.buildProcess();
				}
			}catch(e){alert(e);}
		};
		
		this.req.onError = function(status, statusText){
			if(size == 0){
				Crawler.onError(status, statusText, bookmark);
				Crawler.buildProcess();
			}else {
				Crawler.redirectProcess(bookmark);
			}
		};
		
		this.req.onRedirect = function(status, locations){
			if(size == 0){
				Crawler.onError(-406, 'Unable to redirect: ' + Book.getURI(bookmark), bookmark);
				Crawler.buildProcess();
			}else {
				Crawler.redirectProcess(bookmark);
			}
		};
		
		this.req.send(uri);
	},
	
	handleRedirect : function(status, locations, bookmark){
		switch(status){
			case 300 : 
				this.redirections = locations;
				this.redirectProcess(bookmark);
				return;
			case 301 : 
				this.redirections = locations;
				this.redirectProcess(bookmark);
				return;
			case 302 : 
				this.redirections = locations;
				this.redirectProcess(bookmark);
				return;
			case 304 : 
				this.redirections = locations;
				this.redirectProcess(bookmark);
				return;
			default : {
				//should never happen
				this.logger.fatal("Unknown status("+status+") at "+Book.getURI(bookmark), this, arguments.callee);
				this.buildProcess();
			}
		}
	},
	
	finalize : function(responseText, mimeType, charSet, bookmark){
		this.onComplete(responseText, mimeType, charSet, bookmark);
		this.logger.info(Book.getURI(bookmark), this, arguments.callee);
		this.buildProcess();
	}
	
};
